'use strict';

// import { Maybe, Parameters, ReturnType } from './declarations';

/**
 * Does nothing
 */
function noOp() {}

/**
 * Get union of two sets.
 * @param setA
 * @param setB
 * @returns union of two sets
 */
function union(setA, setB) {
  return new Set([...setA, ...setB]);
}

/**
 * Get intersection of two sets.
 * @param setA
 * @param setB
 * @returns intersection of two sets
 */
function intersection(setA, setB) {
  return new Set([...setA].filter(x => setB.has(x)));
}

/**
 * Get difference of two sets.
 * @param setA
 * @param setB
 * @returns difference between two sets
 */
function difference(setA, setB) {
  return new Set([...setA].filter(x => !setB.has(x)));
}

/**
 * Identifies if value is NaN.
 * @param val
 */
function isNaN(val) {
  if (typeof val === 'number') {
    return Number.isNaN(val);
  }
  return false;
}

/**
 * Identifies if value is not defined.
 * undefined, null and NaN are not considered as defined.
 * @param val
 */
function isNotDefined(val) {
  return val === undefined || val === null || isNaN(val);
}

/**
 * Identifies if value is defined.
 * @param val
 */
function isDefined(val) {
  return !isNotDefined(val);
}

/**
 * Identifies if value is falsy.
 * undefined, null, NaN and false are considered as false value.
 * @param val
 * @param override list of additional values that are considered false values
 */
function isFalsy(val) {
  let override = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  return isNotDefined(val) || val === false || override.includes(val);
}

/**
 * Identifies if value is truthy.
 * @param val
 * @param override list of additional values that are considered true values
 */
function isTruthy(val) {
  let override = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  return !isFalsy(val) || override.includes(val);
}

/**
 * Identifies if value is falsy string.
 * undefined, null, NaN, false, '' are considered as false value.
 * @param val
 */
function isFalsyString(val) {
  return isFalsy(val, ['']);
}

/**
 * Identifies if value is truthy string.
 * @param val
 */
function isTruthyString(val) {
  return !isFalsyString(val);
}

// eslint-disable-next-line @typescript-eslint/no-explicit-any

// eslint-disable-next-line @typescript-eslint/no-explicit-any
function isFn(value) {
  return typeof value === 'function';
}

/**
 * Resolves if first argument is function with other arguments as argument to
 * user supplied function.
 * @param variable either a function to resolive or any value
 * @param args arguments for user supplied function
 * @returns value resolved from user supplied function
 */
// eslint-disable-next-line @typescript-eslint/no-explicit-any, max-len
// eslint-disable-next-line @typescript-eslint/no-explicit-any, max-len
function resolve(variable) {
  if (isFn(variable)) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    const value = variable(...args);
    return value;
  }
  return variable;
}

/**
 * Get color from a number.
 * @param code
 * @returns color derived from hash
 */
function getHexFromCode(code) {
  // eslint-disable-next-line no-bitwise
  const color = (code & 0x00FFFFFF).toString(16).toUpperCase();

  // FIXME: use padStart
  const hex = '00000'.substring(0, 6 - color.length) + color;
  return `#${hex}`;
}

/**
 * Get hash from a string.
 * @param str any string
 * @returns hash derived from string
 */
function getHashFromString(str) {
  let hash = 0;
  for (let i = 0; i < str.length; i += 1) {
    hash = str.charCodeAt(i) + ((hash << 5) - hash); // eslint-disable-line
  }

  return hash;
}

/**
 * Get color from string.
 * @param str any string
 */
function getHexFromString(str) {
  return getHexFromCode(getHashFromString(str));
}

/**
 * Get hex color from rgb color.
 * @param color
 */
function getHexFromRgbRaw(color) {
  const [r, g, b] = color;
  // eslint-disable-next-line no-bitwise
  return `#${((r << 16) + (g << 8) + b).toString(16).padStart(6, '0')}`;
}

/**
 * Get hex color from rgb color.
 * @param color
 */
function getHexFromRgb(color) {
  const result = /^rgb\((\d+),\s*(\d+),\s*(\d+)/i.exec(color);
  if (!result) {
    return undefined;
  }
  const [, rs, gs, bs] = result;
  if (!rs || !gs || !bs) {
    return undefined;
  }
  const r = parseInt(rs, 10);
  const g = parseInt(gs, 10);
  const b = parseInt(bs, 10);
  return getHexFromRgbRaw([r, g, b]);
}

/**
 * Get rgb color from hex color.
 * @param color
 */
function getRgbRawFromHex(color) {
  const result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(color);
  if (!result) {
    return undefined;
  }
  const [, rs, gs, bs] = result;
  if (!rs || !gs || !bs) {
    return undefined;
  }
  return [parseInt(rs, 16), parseInt(gs, 16), parseInt(bs, 16)];
}

/**
 * Get rgb color from hex color 'rgb(255, 255, 255)'.
 * @param hex hex color '#ffffff'
 */
function getRgbFromHex(hex) {
  const color = getRgbRawFromHex(hex);
  if (isNotDefined(color)) {
    return undefined;
  }
  const [r, g, b] = color;
  return `rgb(${r}, ${g}, ${b})`;
}

/**
 * Convert hex to yiq colorspace and get the luma value
 * @param color
 * @returns yiq luma value
 */
// https://en.wikipedia.org/wiki/YIQ#From_RGB_to_YIQ
function getContrastYIQ(color) {
  const rgbColor = getRgbRawFromHex(color);
  if (!rgbColor) {
    return 0;
  }
  const [r, g, b] = rgbColor;
  const yiq = (r * 299 + g * 587 + b * 114) / (255 * 1000);
  return yiq;
}

/**
 * Identifies if given color is valid color.
 * @param color
 */
function isValidHexColor(color) {
  const colorHex = /(^#[0-9A-F]{6}$)|(^#[0-9A-F]{3}$)/i;
  return colorHex.test(color);
}

/**
 * Get appropriate foreground color for a given background color.
 * @param color color for background
 * @param colorOnLight color for light background
 * @param colorOnDark color for dark background
 * @returns appropriate foreground color
 */
function getColorOnBgColor(color) {
  let colorOnLight = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '#212121';
  let colorOnDark = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '#ffffff';
  return getContrastYIQ(color) > 0.5 ? colorOnLight : colorOnDark;
}

/**
 * Interpolate two rgb colors.
 * @param color1
 * @param color2
 * @param factor decides where the interpolated color lies.
 * @returns interpolated rgb color
 */
// Taken from the awesome ROT.js roguelike dev library at
// https://github.com/ondras/rot.js
function interpolateRgb(color1, color2) {
  let factor = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0.5;
  const result = [Math.round(factor * (color2[0] - color1[0]) + color1[0]), Math.round(factor * (color2[1] - color1[1]) + color1[1]), Math.round(factor * (color2[2] - color1[2]) + color1[2])];
  return result;
}

function comparison(extractor, comparisonFunc) {
  return function (x, y) {
    let direction = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;
    const a = isDefined(x) ? extractor(x) : undefined;
    const b = isDefined(y) ? extractor(y) : undefined;
    if (a === b) {
      return 0;
    }
    if (isNotDefined(a)) {
      return direction * 1;
    }
    if (isNotDefined(b)) {
      return direction * -1;
    }
    return direction * comparisonFunc(a, b);
  };
}

// NOTE: func is never called for boolean
const compareBoolean = comparison(x => x, (a, b) => Number(a) - Number(b));
const compareString = comparison(x => x, (a, b) => a.localeCompare(b));
const compareNumber = comparison(x => x, (a, b) => a - b);
const compareDate = comparison(x => x, (a, b) => {
  const dateA = new Date(a);
  const dateB = new Date(b);
  return dateA.getTime() - dateB.getTime();
});
const compareStringAsNumber = comparison(x => Number(x), (a, b) => a - b);
const compareLength = comparison(x => x.length, (a, b) => a - b);
const compareStringByWordCount = comparison(x => x.split(/\s+/).length, (a, b) => a - b);
const compareStringSearch = (x, y, z, d) => {
  if (!z) {
    return 0;
  }
  return comparison(a => a, (a, b) => {
    const searchText = z.toLowerCase();
    const firstIndex = a.toLowerCase().indexOf(searchText);
    const secondIndex = b.toLowerCase().indexOf(searchText);
    if (firstIndex === secondIndex) {
      return a.localeCompare(b);
    }
    if (firstIndex === -1 && secondIndex === -1) {
      return a.localeCompare(b);
    }
    if (secondIndex === -1) {
      return -1;
    }
    if (firstIndex === -1) {
      return 1;
    }
    return firstIndex - secondIndex;
  })(x, y, d);
};

/**
 * Concatenate valid string into classname
 * @param props arguments to concatenate
 * @returns final classname
 */
// eslint-disable-next-line import/prefer-default-export, no-underscore-dangle
function _cs() {
  for (var _len = arguments.length, props = new Array(_len), _key = 0; _key < _len; _key++) {
    props[_key] = arguments[_key];
  }
  return props.filter(isTruthyString).join(' ');
}

/**
 * Add padding before string or number
 *
 * @param value
 * @param length total length of the output
 * @param pad character used to pad
 *
 * @remarks
 * Use when String(num).padStart(length, str) not fully supported
 */
function padStart(value, length) {
  let pad = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '0';
  if (isNotDefined(value)) {
    return '';
  }
  const str = String(value);
  return str.length >= length ? str : Array(length - str.length + 1).join(pad) + str;
}

/**
 * Change the first letter of word to uppercase
 *
 * @param str
 */

function capitalize(str) {
  if (isTruthyString(str)) {
    return str.replace(/\b\w/g, l => l.toUpperCase());
  }
  return str;
}

/**
 * Format text, extracted from pdfs, to remove extraneous spaces
 *
 * @param text
 */
function formatPdfText(text) {
  return text
  // Tab to space
  .replace(/\t/gi, ' ')
  // Anything except ascii TODO: support accents and other utf-8
  // text = text.replace(/[^\x00-\x7F]/g, "");
  // Single line break to space
  .replace(/([^\n])[ \t]*\n[ \t]*(?!\n)/gi, '$1 ')
  // Multiple spaces to single
  .replace(/ +/gi, ' ')
  // More than 3 line breaks to just 3 line breaks
  .replace(/\n\s*\n\s*(\n\s*)+/gi, '\n\n\n')
  // This weird -? text combo to just -
  .replace(/-\?/gi, '-').trim();
}

/**
 * Get rating for content in string
 *
 * @param content
 * @param str
 */
function getRatingForContentInString(content, str) {
  // FIXME: the behavior needs to be tested again
  if (isFalsyString(content) || isFalsyString(str)) {
    return -1;
  }
  return content.toLowerCase().indexOf(str.toLowerCase());
}

/**
 * Get random string
 *
 * @param length length of the random string
 * @param mixedCase if uppercase alphabets are to be included
 */
function randomString() {
  let length = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 16;
  let mixedCase = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
  let text = '';
  const possible = mixedCase ? 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789' : 'abcdefghijklmnopqrstuvwxyz0123456789';
  for (let i = 0; i < length; i += 1) {
    text += possible.charAt(Math.floor(Math.random() * possible.length));
  }
  return text;

  /*
   * -------------------------------------------------
   * better Algorithm, but not supported by enzyme :(
   * -------------------------------------------------
   const randomValues = new Uint8Array(length);
  window.crypto.getRandomValues(randomValues);
  return Array.from(randomValues, v => v.toString(36)).join('').substring(0, 8);
  */
}

/**
 * Convert camel case to space separated words
 *
 * @param str
 * @param separator
 *
 * @remarks
 * Only support alphabets (not numerals)
 */
const reOne = /([a-z])([A-Z])([A-Z])/g;
const reTwo = /([a-z])([A-Z])/g;
const reThree = /([A-Z])([A-Z])([a-z])/g;
function camelToNormal(str) {
  let separator = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : ' ';
  if (isNotDefined(str)) {
    return str;
  }
  return str.replace(reOne, (_, one, two, three) => `${one}${separator}${two}${three}`).replace(reTwo, (_, one, two) => `${one}${separator}${two.toLowerCase()}`).replace(reThree, (_, one, two, three) => `${one}${separator}${two.toLowerCase()}${three}`);
}

/**
 * Convert camel case to snake case
 *
 * @param str
 *
 * @remarks
 * Only support alphabets (not numerals)
 */

function camelToSnake(str) {
  if (isNotDefined(str)) {
    return str;
  }
  const value = camelToNormal(str, '_');
  return value.charAt(0).toLowerCase() + value.slice(1);
}

/**
 * Convert camel case to kebab case
 *
 * @param str
 *
 * @remarks
 * Only support alphabets (not numerals)
 */

function camelToKebab(str) {
  if (isNotDefined(str)) {
    return str;
  }
  const value = camelToNormal(str, '-');
  return value.charAt(0).toLowerCase() + value.slice(1);
}

/**
 * Split sentence
 *
 * @param str
 */
function splitInWhitespace(str) {
  if (isNotDefined(str)) {
    return [];
  }
  return str.match(/\S+/g) ?? [];
}

/**
 * Convert all whitespaces into single space
 *
 * @param str
 */

function trimWhitespace(str) {
  if (isNotDefined(str)) {
    return str;
  }
  return splitInWhitespace(str).join(' ');
}

/**
 * Identify if shortText is inside longText
 *
 * @param longText
 * @param shortText
 *
 * @remarks
 * The match is case-insensitive
 *
 */
function caseInsensitiveSubmatch(longText, shortText) {
  if (isNotDefined(longText) || isNotDefined(shortText)) {
    return false;
  }
  return String(longText).trim().toLowerCase().includes(String(shortText).trim().toLowerCase());
}

// FIXME: do not use date library

const MONTHS = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
const DAYS = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
var ValueType = /*#__PURE__*/function (ValueType) {
  ValueType["time"] = "time";
  ValueType["date"] = "date";
  return ValueType;
}(ValueType || {});
function getStartAndEnd(format, matches) {
  let minStart;
  let maxEnd;
  matches.forEach(val => {
    const start = format.indexOf(val);
    if (start === -1) {
      return;
    }
    const end = start + val.length;
    if (minStart === undefined || start < minStart) {
      minStart = start;
    }
    if (maxEnd === undefined || end > maxEnd) {
      maxEnd = end;
    }
  });
  return {
    start: minStart,
    end: maxEnd
  };
}

/**
 * Break timedate format into time unit, date unit and other unit.
 *
 * @remarks
 * The date unit are: yyyy, yy, MMM, MM, dd, EEE.
 * The time unit are: hh, mm, ss, aaa.
 * Anything can come in between date and time unit and they will be treated as separator.
 * @param format format for datetime
 */
function breakFormat(format) {
  const {
    start: dateStart,
    end: dateEnd
  } = getStartAndEnd(format, ['yyyy', 'yy', 'MMM', 'MM', 'dd', 'EEE']);
  const {
    start: timeStart,
    end: timeEnd
  } = getStartAndEnd(format, ['hh', 'mm', 'ss', 'aaa']);
  const formatEnd = format.length;
  const getSubFormat = (start, end) => {
    if (start === undefined || end === undefined) {
      return '';
    }
    return format.substring(start, end);
  };
  if (dateStart === undefined && timeStart === undefined) {
    return [{
      value: format
    }].filter(a => a.value !== '');
  }
  if (dateStart === undefined) {
    return [{
      value: getSubFormat(0, timeStart)
    }, {
      type: ValueType.time,
      value: getSubFormat(timeStart, timeEnd)
    }, {
      value: getSubFormat(timeEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  if (timeStart === undefined) {
    return [{
      value: getSubFormat(0, dateStart)
    }, {
      type: ValueType.date,
      value: getSubFormat(dateStart, dateEnd)
    }, {
      value: getSubFormat(dateEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  if (dateStart < timeStart) {
    return [{
      value: getSubFormat(0, dateStart)
    }, {
      type: ValueType.date,
      value: getSubFormat(dateStart, dateEnd)
    }, {
      value: getSubFormat(dateEnd, timeStart)
    }, {
      type: ValueType.time,
      value: getSubFormat(timeStart, timeEnd)
    }, {
      value: getSubFormat(timeEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  return [{
    value: getSubFormat(0, timeStart)
  }, {
    type: ValueType.time,
    value: getSubFormat(timeStart, timeEnd)
  }, {
    value: getSubFormat(timeEnd, dateStart)
  }, {
    type: ValueType.date,
    value: getSubFormat(dateStart, dateEnd)
  }, {
    value: getSubFormat(dateEnd, formatEnd)
  }].filter(a => a.value !== '');
}

/**
 * Populate the format list with date
 *
 * @param formatList the list of datetime units
 * @param date
 */
function populateFormat(formatList, date) {
  return formatList.map(format => {
    if (format.type === ValueType.date) {
      const year = date.getFullYear();
      const month = date.getMonth();
      const day = date.getDate();
      const weekName = DAYS[date.getDay()];
      const newFormat = {
        ...format
      };
      newFormat.value = newFormat.value.replace('yyyy', String(year)).replace('yy', String(year % 100)).replace('MMM', MONTHS[month]).replace('MM', padStart(month + 1, 2)).replace('EEE', weekName).replace('dd', padStart(day, 2));
      return newFormat;
    }
    if (format.type === ValueType.time) {
      const ttIndex = format.value.indexOf('aaa');
      const originalHour = date.getHours();
      const hour = ttIndex >= 0 ? (originalHour - 1) % 12 + 1 : originalHour;
      const amPm = originalHour >= 12 ? 'PM' : 'AM';
      const minute = date.getMinutes();
      const second = date.getSeconds();
      const newFormat = {
        ...format
      };
      newFormat.value = newFormat.value.replace('hh', padStart(hour, 2)).replace('mm', padStart(minute, 2)).replace('ss', padStart(second, 2)).replace('aaa', amPm);
      return newFormat;
    }
    return format;
  });
}

/**
 * Format date
 *
 * @param date
 * @param format format for datetime
 * @returns formatted date
 */
function formatDateToString(date, format) {
  return populateFormat(breakFormat(format), date).map(e => e.value).join('');
}

/**
 * Set hour, minute and second to zero in given datetime
 *
 * @param date timestamp or a date string
 * @returns timestamp with hour, minute and second set to zero
 */
function getDate(datetime) {
  const today = new Date(datetime);
  today.setHours(0, 0, 0, 0);
  return today.getTime();
}

/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
function getDifferenceInDays(a, b) {
  const dateA = getDate(a);
  const dateB = getDate(b);
  return (dateA - dateB) / (1000 * 60 * 60 * 24);
}

/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
function getDateDifferenceHumanReadable(a, b) {
  const diff = getDifferenceInDays(a, b);
  if (diff === 0) {
    return 'Today';
  }
  if (diff === -1) {
    return 'Yesterday';
  }
  if (diff === 1) {
    return 'Tomorrow';
  }
  if (diff < 0) {
    return `${-diff} days ago`;
  }
  return `After ${diff} days`;
}

/**
 * Get number of days in certain year and month
 *
 * @param year
 * @param month
 * @returns number of days
 *
 * @remarks
 * Month starts from 1 (not zero)
 */
function getNumDaysInMonthX(year, month) {
  if (isTruthy(year) && isTruthy(month)) {
    return new Date(year, month, 0).getDate();
  }
  return 32;
}

/**
 * Get number of days in certain year and month from a date
 *
 * @param date
 * @returns number of days
 */
function getNumDaysInMonth(date) {
  if (date) {
    return getNumDaysInMonthX(date.getFullYear(), date.getMonth() + 1);
  }
  return 32;
}

/**
 * Change date into 'yyyy-MM-dd' string
 *
 * @param date
 */
function encodeDate(date) {
  return formatDateToString(date, 'yyyy-MM-dd');
}

/**
 * Change timestamp or 'yyyy-MM-dd' string into date
 *
 * @param value timestamp or 'yyyy-MM-dd' string
 */

function decodeDate(value) {
  // Let's assume that the value is in local time zone

  if (!value) {
    return undefined;
  }

  // Check if value is timestamp number or ISO string with time information
  // In both case, new Date assumes local time zone
  if (typeof value === 'number' || value.includes('T')) {
    return new Date(value);
  }

  // In case of ISO string with no time information, new Date assumes
  // UTC timezone. So first split it manually, and feed them separately
  // so they will be processed as local time zone.
  const splits = value.split('-');
  return new Date(Number(splits[0]), Number(splits[1]) - 1, Number(splits[2]));
}
const MIN_YEAR = 1990;
/**
 * Identify if ymd value is complete
 *
 * @param ymd
 */
function isDateValuesComplete(ymd) {
  const {
    yearValue,
    monthValue,
    dayValue
  } = ymd;
  // Complete if all values are undefined or none are
  return isTruthy(dayValue) && isTruthy(monthValue) && isTruthy(yearValue) || isFalsy(dayValue) && isFalsy(monthValue) && isFalsy(yearValue);
}

/**
 * Identify problem with ymd value
 *
 * @param ymd
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
function getErrorForDateValues(ymd) {
  if (!isDateValuesComplete(ymd)) {
    return 'Date values incomplete';
  }
  const {
    yearValue,
    monthValue,
    dayValue
  } = ymd;
  if (yearValue < MIN_YEAR) {
    return `Year should be greater than or equal to ${MIN_YEAR}`;
  }
  if (monthValue < 1 || monthValue > 12) {
    return 'Month should be between 1 and 12';
  }
  const maxNumberOfDays = getNumDaysInMonthX(yearValue, monthValue);
  if (dayValue < 1 || dayValue > maxNumberOfDays) {
    return 'Invalid day for this month';
  }
  return undefined;
}
const MIN_HOUR = 0;
const MAX_HOUR = 23;
const MIN_MINUTE = 0;
const MAX_MINUTE = 59;
const MIN_SECOND = 0;
const MAX_SECOND = 59;
/**
 * Identify if hms value is complete
 *
 * @param hms
 */
const isTimeValuesComplete = val => {
  const {
    hourValue,
    minuteValue,
    secondValue
  } = val;
  // Complete if all values are undefined or none are
  return isTruthy(hourValue) && isTruthy(minuteValue) && isTruthy(secondValue) || isFalsy(hourValue) && isFalsy(minuteValue) && isFalsy(secondValue);
};

/**
 * Identify problem with hms value
 *
 * @param hms
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
const getErrorForTimeValues = val => {
  if (!isTimeValuesComplete(val)) {
    return 'Time values incomplete';
  }
  const {
    hourValue,
    minuteValue,
    secondValue
  } = val;
  if (hourValue < MIN_HOUR || hourValue > MAX_HOUR) {
    return `Hour must be between ${MIN_HOUR} and ${MAX_HOUR}`;
  }
  if (minuteValue < MIN_MINUTE || minuteValue > MAX_MINUTE) {
    return `Minute must be between ${MIN_MINUTE} and ${MAX_MINUTE}`;
  }
  if (secondValue < MIN_SECOND || secondValue > MAX_SECOND) {
    return `Second must be between ${MIN_SECOND} and ${MAX_SECOND}`;
  }
  return undefined;
};

/**
 * Get nth node from a linked list.
 * @param node the first node of the linked list
 * @param n the nth node after the first node
 * @param selector get next node from current node
 *
 * @returns the nth node
 */ // eslint-disable-next-line import/prefer-default-export
function getLinkedListNode(node, n, selector) {
  let newNode = node;
  for (let i = 0; i < n; i += 1) {
    newNode = selector(newNode, i);
    if (isNotDefined(newNode)) {
      break;
    }
  }
  return newNode;
}

/**
 * Transform list to object
 * @param list
 * @param keySelector get key from item in list
 * @param modifier get new value from item in list
 */

function listToMap(list, keySelector, modifier) {
  if (isNotDefined(list)) {
    return undefined;
  }
  const value = list.reduce((acc, elem, i) => {
    const key = keySelector(elem, i);
    acc[key] = modifier ? modifier(elem, key, i, acc) : elem;
    return acc;
  }, {});
  return value;
}

/**
 * Transform object to list
 * @param obj
 * @param modifier get new value from item in object
 */

function mapToList(obj, modifier) {
  if (isNotDefined(obj)) {
    return undefined;
  }
  return Object.keys(obj).reduce((acc, key, i) => {
    const elem = obj[key];
    acc.push(modifier ? modifier(elem, key, i, acc) : elem);
    return acc;
  }, []);
}

/**
 * Transform object to object
 * @param obj
 * @param keySelector get new key for item in object
 * @param modifier get new value for item in object
 */

function mapToMap(obj, keySelector, modifier) {
  if (isNotDefined(obj)) {
    return undefined;
  }
  const value = Object.keys(obj).reduce((acc, k, i) => {
    const elem = obj[k];
    const key = keySelector ? keySelector(k, elem) : k;
    acc[key] = modifier ? modifier(elem, k, i, acc) : elem;
    return acc;
  }, {});
  return value;
}

/**
 * Group list into object of list.
 * Items in list are grouped by key returned by keySelector
 * @param list
 * @param keySelector get key for item in list
 * @param modifier get new value for item in list
 */

function listToGroupList(list, keySelector, modifier, groupModifier) {
  if (isNotDefined(list)) {
    return undefined;
  }
  const val = list.reduce((acc, elem, i) => {
    const key = keySelector(elem, i);
    const value = modifier ? modifier(elem, key, i, acc) : elem;
    const group = acc[key];
    if (group) {
      group.push(value);
    } else {
      acc[key] = [value];
    }
    return acc;
  }, {});
  if (groupModifier) {
    return Object.keys(val).reduce((acc, key) => {
      const value = val[key];
      if (value) {
        acc[key] = groupModifier(value, key);
      }
      return acc;
    }, {});
  }
  return val;
}

/**
 * Identify if two list are the same
 *
 * @param list1
 * @param list2
 *
 * @remarks
 * The function assumes that the elements aren't mutated
 */
function isListEqual(list1, list2) {
  return list1.length === list2.length && list1.every((d, i) => d === list2[i]);
}

/**
 * Get a random item from the list
 *
 * @param items
 */
function getRandomFromList(items) {
  return items[Math.floor(Math.random() * items.length)];
}

/**
 * Get an element around certain index in a list
 *
 * @param list
 * @param index
 */
function getElementAround(list, index) {
  if (index + 1 < list.length) {
    return list[index + 1];
  }
  if (index - 1 >= 0) {
    return list[index - 1];
  }
  return undefined;
}

/**
 * Get a defined element around certain index in a list
 *
 * @param list
 * @param index
 */
function getDefinedElementAround(list, currentIndex) {
  let i;
  let j;
  for (i = currentIndex - 1; i >= 0; i -= 1) {
    if (list[i] !== undefined) {
      break;
    }
  }
  for (j = currentIndex + 1; j < list.length; j += 1) {
    if (list[j] !== undefined) {
      break;
    }
  }
  if (i < 0 && j >= list.length) {
    // console.warn('none');
    return undefined;
  }
  if (i < 0 && j < list.length) {
    // console.warn('none on left');
    return list[j];
  }
  if (j >= list.length && i >= 0) {
    // console.warn('none on right');
    return list[i];
  }
  const iDist = currentIndex - i;
  const jDist = j - currentIndex;
  if (jDist > iDist) {
    return list[i];
  }
  return list[j];
}

/**
 * Get a duplicate count for each element in a list
 *
 * @param list
 * @param keySelector get key from an element
 *
 * @returns map of element key and duplicate count
 *
 * @remarks
 * The map only includes element for which there is a duplicate
 */

function getDuplicates(list, keySelector) {
  if (!list) {
    return undefined;
  }
  const counts = listToMap(list, keySelector, (_, key, __, acc) => {
    const value = acc[key];
    return isDefined(value) ? value + 1 : 1;
  });
  return Object.keys(counts).filter(key => counts[key] > 1);
}

/**
 * Find difference between two list
 *
 * @param listA
 * @param listB
 * @param keySelector get key from an element
 *
 * @returns list of added, modified, removed and unmodified elements
 */
function findDifferenceInList(listA, listB, keySelector) {
  const modified = [];
  const added = [];
  const removed = [];
  const unmodified = [];
  const mapA = listToMap(listA, keySelector, e => e);
  listB.forEach((elem, i) => {
    const key = keySelector(elem, i);
    if (isNotDefined(mapA[key])) {
      added.push(elem);
    } else if (mapA[key] !== elem) {
      modified.push({
        old: mapA[key],
        new: elem
      });
    } else {
      unmodified.push(elem);
    }
  });
  const mapB = listToMap(listB, keySelector, e => e);
  listA.forEach((elem, i) => {
    const key = keySelector(elem, i);
    if (isNotDefined(mapB[key])) {
      removed.push(elem);
    }
  });
  return {
    added,
    modified,
    removed,
    unmodified
  };
}

/**
 * Find unique items from a list
 *
 * @param list
 * @param getItemHash method to get id to check uniqueness
 *
 * @remarks
 * If getItemHash is not supplied, comparision is done by casting items in list
 * to string
 */

function unique(list, getItemHash) {
  if (isNotDefined(list)) {
    return undefined;
  }
  const memory = {};
  const arrWithUnique = [];
  list.forEach(item => {
    const hash = getItemHash ? getItemHash(item) : JSON.stringify(item);
    if (!memory[hash]) {
      memory[hash] = true;
      arrWithUnique.push(item);
    }
  });
  // NOTE: just return original array if all elements were unique
  if (list.length === arrWithUnique.length) {
    return list;
  }
  return arrWithUnique;
}
function max(list, getNumericValue) {
  if (!list || list.length <= 0) {
    return undefined;
  }
  const values = list.reduce((acc, item) => {
    const {
      maxValue
    } = acc;
    const myValue = getNumericValue(item);
    return isDefined(myValue) && (isNotDefined(maxValue) || myValue > maxValue) ? {
      maxValue: myValue,
      maxItem: item
    } : acc;
  }, {
    maxItem: undefined,
    maxValue: undefined
  });
  return values.maxItem;
}
function min(list, getNumericValue) {
  if (!list || list.length <= 0) {
    return undefined;
  }
  const values = list.reduce((acc, item) => {
    const {
      minValue
    } = acc;
    const myValue = getNumericValue(item);
    return isDefined(myValue) && (isNotDefined(minValue) || myValue < minValue) ? {
      minValue: myValue,
      minItem: item
    } : acc;
  }, {
    minItem: undefined,
    minValue: undefined
  });
  return values.minItem;
}

/**
 * Check if new object should be set
 * and should the message of overriden shown
 *
 * @param oldVersionId
 * @param newVersionId
 */
// eslint-disable-next-line import/prefer-default-export
function checkVersion(oldVersionId, newVersionId) {
  return {
    shouldSetValue: isFalsy(oldVersionId) || oldVersionId < newVersionId,
    isValueOverriden: isTruthy(oldVersionId) && oldVersionId < newVersionId
  };
}

// TODO: get methods from Numeral
/**
 * Get comma separated number
 * @param num
 * @param separator thousand separator
 * @param decimalSeparator decimal separator
 */
function addSeparator(num) {
  let separator = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : ',';
  let decimalSeparator = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '.';
  if (isNotDefined(num)) {
    return num;
  }
  const [before = '', after = ''] = String(num).split(decimalSeparator);
  let x1 = before;
  const rgx = /(\d+)(\d{3})/;
  while (rgx.test(x1)) {
    x1 = x1.replace(rgx, `$1${separator}$2`);
  }
  const x2 = after !== '' ? `${decimalSeparator}${after}` : '';
  return x1 + x2;
}

/**
 * Get normalized number
 * @param num
 * @param lang language
 */
function formattedNormalize(num) {
  const mapping = [{
    suffix: 'B',
    value: 1000000000
  }, {
    suffix: 'M',
    value: 1000000
  }, {
    suffix: 'K',
    value: 1000
  }];
  const rule = mapping.find(n => num >= n.value);
  if (isNotDefined(rule)) {
    return {
      number: num
    };
  }
  const {
    suffix,
    value
  } = rule;
  return {
    number: num / value,
    normalizeSuffix: suffix
  };
}

/**
 * Identifies if item is a list.
 * @param item anything
 */
function isList(item) {
  return Array.isArray(item);
}

/**
 * Identifies if item is an object.
 * `null` and `list` are not considered as object.
 * @param item anything
 */
function isObject(item) {
  return item !== null && typeof item === 'object' && !isList(item);
}

/**
 * Identifies if item is an integer.
 * Float values are not not considered as integer.
 * @param item anything
 */
function isInteger(value) {
  return typeof value === 'number' && value % 1 === 0;
}
const basicTypes = ['Boolean', 'Number', 'String', 'Function', 'Array', 'Date', 'RegExp', 'Object', 'Error', 'Symbol'];

// Mapping to hold javascript class to custom type
const classToType = basicTypes.reduce((acc, type) => ({
  ...acc,
  [`[object ${type}]`]: type.toLowerCase()
}), {});

/**
 * Get type of an item.
 * Valid types: null, boolean, number, string, function, array, date, regexp, object, error, symbol
 * @param item anything
 * @returns type of that item
 */
const typeOf = item => {
  if (item == null) {
    return 'null';
  }
  const simpleType = typeof item;
  return simpleType === 'object' || simpleType === 'function' ? classToType[classToType.toString.call(item)] ?? 'object' : simpleType;
};

/**
 * Identifies if string is a valid email.
 * @param item any string
 */
function isValidEmail(value) {
  const re = /^(([^<>()[\]\\.,;:\s@"]+(\.[^<>()[\]\\.,;:\s@"]+)*)|(".+"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/;
  return re.test(value);
}

/**
 * Identifies if string is a valid url.
 * @param item any string
 */
function isValidUrl(value) {
  //
  // Regular Expression for URL validation
  //
  // Author: Diego Perini
  // Updated: 2010/12/05
  // License: MIT
  //
  // Copyright (c) 2010-2013 Diego Perini (http://www.iport.it)
  //
  // Permission is hereby granted, free of charge, to any person
  // obtaining a copy of this software and associated documentation
  // files (the "Software"), to deal in the Software without
  // restriction, including without limitation the rights to use,
  // copy, modify, merge, publish, distribute, sublicense, and/or sell
  // copies of the Software, and to permit persons to whom the
  // Software is furnished to do so, subject to the following
  // conditions:
  //
  // The above copyright notice and this permission notice shall be
  // included in all copies or substantial portions of the Software.
  //
  // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  // EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
  // OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  // NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  // HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
  // WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  // FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  // OTHER DEALINGS IN THE SOFTWARE.
  //
  // the regular expression composed & commented
  // could be easily tweaked for RFC compliance,
  // it was expressly modified to fit & satisfy
  // these test for an URL shortener:
  //
  //   http://mathiasbynens.be/demo/url-regex
  //
  // Notes on possible differences from a standard/generic validation:
  //
  // - utf-8 char class take in consideration the full Unicode range
  // - TLDs have been made mandatory so single names like "localhost" fails
  // - protocols have been restricted to ftp, http and https only as requested
  //
  // Changes:
  //
  // - IP address dotted notation validation, range: 1.0.0.0 - 223.255.255.255
  //   first and last IP address of each class is considered invalid
  //   (since they are broadcast/network addresses)
  //
  // - Added exclusion of private, reserved and/or local networks ranges
  //
  // - Made starting path slash optional (http://example.com?foo=bar)
  //
  // - Allow a dot (.) at the end of hostnames (http://example.com.)
  //

  const re = new RegExp('^'
  // protocol identifier
  + '(?:(?:https?|ftp)://)'
  // user:pass authentication
  + '(?:\\S+(?::\\S*)?@)?' + '(?:'
  // IP address exclusion
  // private & local networks
  + '(?!(?:10|127)(?:\\.\\d{1,3}){3})' + '(?!(?:169\\.254|192\\.168)(?:\\.\\d{1,3}){2})' + '(?!172\\.(?:1[6-9]|2\\d|3[0-1])(?:\\.\\d{1,3}){2})'
  // IP address dotted notation octets
  // excludes loopback network 0.0.0.0
  // excludes reserved space >= 224.0.0.0
  // excludes network & broacast addresses
  // (first & last IP address of each class)
  + '(?:[1-9]\\d?|1\\d\\d|2[01]\\d|22[0-3])' + '(?:\\.(?:1?\\d{1,2}|2[0-4]\\d|25[0-5])){2}' + '(?:\\.(?:[1-9]\\d?|1\\d\\d|2[0-4]\\d|25[0-4]))' + '|'
  // host name
  + '(?:(?:[_a-z\\u00a1-\\uffff0-9]-*)*[a-z\\u00a1-\\uffff0-9]+)'
  // domain name
  + '(?:\\.(?:[a-z\\u00a1-\\uffff0-9]-*)*[a-z\\u00a1-\\uffff0-9]+)*'
  // TLD identifier
  + '(?:\\.(?:[a-z\\u00a1-\\uffff]{2,}))'
  // TLD may end with dot
  + '\\.?' + ')'
  // port number
  + '(?::\\d{2,5})?'
  // resource path
  + '(?:[/?#]\\S*)?' + '$', 'i');
  return re.test(value);
}

/**
 * Pick only certain keys from an object
 * @param obj
 * @param keys
 */
const pick = (obj, keys) => keys.reduce((acc, key) => ({
  ...acc,
  [key]: obj[key]
}), {});

/**
 * Remove a certain key from an object
 * @param obj
 * @param key
 */
function removeKey(obj, key) {
  if (isNotDefined(obj) || !(key in obj)) {
    return obj;
  }
  const newObj = {
    ...obj
  };
  // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
  delete newObj[key];
  return newObj;
}

/**
 * Get key of the first element in an object
 * @param obj
 * @param value
 */
function getFirstKeyByValue(obj, value) {
  return Object.keys(obj).find(key => obj[key] === value);
}

/**
 * Identify if object has no data
 * @param obj
 * @param invalids list of values that aren't considered as data
 */
function doesObjectHaveNoData(obj) {
  let invalids = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
  // NOTE: null and NaN are truthy values here
  if (obj === undefined) {
    return true;
  }
  if (isList(obj)) {
    if (obj.length <= 0) {
      return true;
    }
    return obj.every(e => doesObjectHaveNoData(e, invalids));
  }
  if (isObject(obj)) {
    if (Object.keys(obj).length <= 0) {
      return true;
    }
    return Object.keys(obj).every(key => doesObjectHaveNoData(obj[key], invalids));
  }
  return invalids.includes(obj);
}

// FIXME: remove this
/**
 * Fill parameters in route url
 * @param route route template
 * @param params map of values
 */
function reverseRoute(route, params) {
  const paths = route.split('/');
  for (let i = 0; i < paths.length; i += 1) {
    let path = paths[i];
    if (path && path.length > 0 && path.startsWith(':')) {
      path = path.substring(1);
      let param;

      // optional parameter
      if (path.endsWith('?')) {
        param = params[path.replace('?', '')];

        // omit if value not supplied
        if (!param) {
          paths.splice(i, 1);
        } else {
          paths[i] = String(param);
        }
      } else {
        param = params[path];
        if (!param) {
          // eslint-disable-next-line no-console
          console.error(`Value for param ${path} not supplied`);
        }
        paths[i] = String(param);
      }
    }
  }
  return paths.join('/');
}

/**
 * Identify if param is required in certain route
 * @param route route template
 * @param param parameter name
 */
function isParamRequired(route, param) {
  const re = new RegExp(`:${param}([^\\?]|$)`);
  return route.search(re) !== -1;
}

/**
 * Get `a` modulo `b`
 *
 * @param a dividend
 * @param b divisor
 * @remarks
 * Return a positive modulo
 */
function modulo(a, b) {
  const value = a % b;
  if (value < 0) {
    return b + value;
  }
  // NOTE: -0 is not less than 0, so we need to take absolute value
  return Math.abs(value);
}

/**
 * Get sum of all elements in a list
 *
 * @param list of numbers
 *
 * @remarks
 * Return a positive modulo
 */
function sum(values) {
  return values.reduce((acc, b) => acc + b, 0);
}

/**
 * Get mean of all elements in a list
 *
 * @param list of numbers
 */
function mean(values) {
  if (values.length <= 0) {
    return 0;
  }
  return sum(values) / values.length;
}

/**
 * Get median of all elements in a list
 *
 * @param list of numbers
 */
function median(values) {
  if (values.length <= 0) {
    return undefined;
  }
  const sortedValues = [...values].sort();
  const centerIndex = Math.floor(sortedValues.length / 2);
  if (sortedValues.length % 2 !== 0) {
    return sortedValues[centerIndex];
  }
  const firstValue = sortedValues[centerIndex - 1];
  const secondValue = sortedValues[centerIndex];
  if (isDefined(firstValue) && isDefined(secondValue)) {
    return (firstValue + secondValue) / 2.0;
  }
  return firstValue ?? secondValue;
}

// min, max, output

/**
 * Get the bucket for a given number
 *
 * @param value
 * @param buckets
 */
function bucket(value, buckets) {
  const currentBucket = buckets.find(_ref => {
    let [min, max] = _ref;
    return value >= min && value < max;
  });
  return currentBucket && currentBucket[2];
}

/**
 * Get number bounded on two sides
 *
 * @param value
 * @param a upper or lower bound
 * @param b upper or lower bound
 */
function bound(value, a, b) {
  const min = Math.min(a, b);
  const max = Math.max(a, b);
  return Math.max(min, Math.min(max, value));
}

/**
 * Get normalized value for a number
 *
 * @param value
 * @param max max value
 * @param min min value
 */
function normalize(value, max, min) {
  return (value - min) / (max - min);
}

/**
 * Get list of numbers from `start` to `end`
 *
 * @param start number from where list starts
 * @param end number where list ends
 */
function getNumbers(start, end) {
  const list = [];
  for (let i = start; i < end; i += 1) {
    list.push(i);
  }
  return list;
}

/**
 * Get trigrams from a sentence
 * @param sentence
 */
function getTrigrams(sentence) {
  const newWord = sentence.trim().toLowerCase().replace(/[^a-zA-Z0-9]/g, ' ').replace(/\s+/g, '  ');
  const formattedWord = `  ${newWord} `;
  const trigrams = [];
  for (let i = 0; i < formattedWord.length - 2; i += 1) {
    const trigram = formattedWord.substr(i, 3);
    if (trigram[1] !== ' ' || trigram[2] !== ' ') {
      trigrams.push(trigram);
    }
  }
  return new Set(trigrams);
}

/**
 * Get trigram similarity between two sentences
 * @param foo first string
 * @param bar second string
 */
function getTrigramSimilarity(foo, bar) {
  const fooTrigrams = getTrigrams(foo);
  const barTrigrams = getTrigrams(bar);
  const common = intersection(fooTrigrams, barTrigrams);
  const all = union(fooTrigrams, barTrigrams);
  return common.size / all.size;
}

function isMonthValid(value) {
  return value >= 1 && value <= 12;
}
function prepareDateUnchecked(year, month, day, cal) {
  return {
    y: year,
    m: month,
    d: day,
    cal
  };
}
function isDateEqual(foo, bar) {
  return foo.y === bar.y && foo.m === bar.m && foo.d === bar.d;
}
function isDateLessThan(foo, bar) {
  return foo.y < bar.y || foo.y === bar.y && foo.m < bar.m || foo.y === bar.y && foo.m === bar.m && foo.d < bar.d;
}
function isDateGreaterThan(foo, bar) {
  return foo.y > bar.y || foo.y === bar.y && foo.m > bar.m || foo.y === bar.y && foo.m === bar.m && foo.d > bar.d;
}
function isDateLessThanOrEqual(foo, bar) {
  return isDateEqual(foo, bar) || isDateLessThan(foo, bar);
}
function isDateGreaterThanOrEqual(foo, bar) {
  return isDateEqual(foo, bar) || isDateGreaterThan(foo, bar);
}
function toDateString(date) {
  return `${padStart(date.y, 4)}-${padStart(date.m, 2)}-${padStart(date.d, 2)}`;
}
const minBsDate = prepareDateUnchecked(2000, 9, 17, 'BS');
const maxBsDate = prepareDateUnchecked(2090, 12, 30, 'BS');
const bsData = [[365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 32, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 31, 29, 30, 30, 29, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 32, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 30, 32, 31, 32, 31, 31, 29, 30, 29, 30, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 29, 31], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 29, 30, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 29, 30, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 30], [366, 31, 32, 31, 32, 31, 30, 30, 30, 29, 30, 29, 31], [365, 31, 31, 31, 32, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 31, 32, 31, 31, 31, 30, 29, 30, 29, 30, 30], [365, 31, 32, 31, 32, 31, 30, 30, 30, 29, 29, 30, 30], [366, 31, 31, 32, 32, 31, 30, 30, 30, 29, 30, 30, 30], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 30, 30], [365, 31, 31, 32, 31, 31, 30, 30, 30, 29, 30, 30, 30], [365, 31, 31, 32, 31, 31, 30, 30, 30, 29, 30, 30, 30], [366, 31, 32, 31, 32, 30, 31, 30, 30, 29, 30, 30, 30], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 30, 30], [366, 31, 31, 32, 31, 31, 31, 30, 30, 29, 30, 30, 30], [365, 30, 31, 32, 32, 30, 31, 30, 30, 29, 30, 30, 30], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 30, 30], [365, 30, 32, 31, 32, 31, 30, 30, 30, 29, 30, 30, 30]];
const minAdDate = prepareDateUnchecked(1944, 1, 1, 'AD');
const maxAdDate = prepareDateUnchecked(2034, 4, 12, 'AD');
const adData = [[365, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31], [366, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]];
function isLeapAdYear(year) {
  return year % 4 === 0 && year % 100 !== 0 || year % 400 === 0;
}
function getDaysInYear(year, cal) {
  if (cal === 'AD') {
    return isLeapAdYear(year) ? 366 : 365;
  }
  const minDate = minBsDate;
  const maxDate = maxBsDate;
  const data = bsData;
  if (year < minDate.y || year > maxDate.y) {
    const error = new Error('Year out of bounds');
    throw error;
  }
  const index = year - minDate.y;
  const yearData = data[index];
  if (!yearData) {
    const error = new Error('Year out of bounds');
    throw error;
  }
  return yearData[0];
}
function isLeapYear(year, cal) {
  if (cal === 'AD') {
    return isLeapAdYear(year);
  }
  return getDaysInYear(year, cal) === 366;
}
function getDaysInMonth(year, month, cal) {
  const minDate = cal === 'AD' ? minAdDate : minBsDate;
  const maxDate = cal === 'AD' ? maxAdDate : maxBsDate;
  if (year < minDate.y || year > maxDate.y) {
    const error = new Error('Year out of bounds');
    throw error;
  }
  if (!isMonthValid(month)) {
    const error = new Error('Month out of bounds');
    throw error;
  }
  if (cal === 'AD') {
    const index = isLeapYear(year, cal) ? 1 : 0;
    return adData[index][month];
  }
  const index = year - minDate.y;
  const yearData = bsData[index];
  if (!yearData) {
    const error = new Error('Year out of bounds');
    throw error;
  }
  return yearData[month];
}
function isValidDate(date) {
  const minDate = date.cal === 'AD' ? minAdDate : minBsDate;
  const maxDate = date.cal === 'AD' ? maxAdDate : maxBsDate;
  if (date.y < minDate.y || date.y > maxDate.y) {
    return [false, 'Year out of bounds'];
  }
  if (date.m < 0 || date.m > 12) {
    return [false, 'Month out of bounds'];
  }
  if (date.d < 0 || date.d > getDaysInMonth(date.y, date.m, date.cal)) {
    return [false, 'Day out of bounds'];
  }

  // NOTE: can be inside min/max year but still be out of bounds for conversion
  if (isDateGreaterThan(date, maxDate) || isDateLessThan(date, minDate)) {
    return [false, 'Miti out of bounds'];
  }
  return [true];
}
function prepareDate(year, month, day, cal) {
  const date = prepareDateUnchecked(year, month, day, cal);
  const [valid, errorMessage] = isValidDate(date);
  if (!valid) {
    const error = new Error(errorMessage);
    throw error;
  }
  return date;
}
function getDateDifference(foo, bar) {
  const mitiIsAfterSub = isDateGreaterThan(foo, bar);
  const y = mitiIsAfterSub ? foo.y : bar.y;
  const m = mitiIsAfterSub ? foo.m : bar.m;
  const d = mitiIsAfterSub ? foo.d : bar.d;
  const y1 = !mitiIsAfterSub ? foo.y : bar.y;
  const m1 = !mitiIsAfterSub ? foo.m : bar.m;
  const d1 = !mitiIsAfterSub ? foo.d : bar.d;

  // NOTE: we can take cal from either foo or bar
  const {
    cal
  } = foo;
  const sign = mitiIsAfterSub ? 1 : -1;
  let no = 0;
  if (y !== y1) {
    // If the years differ
    // Get number of days 'no' till end of year
    no = getDaysInMonth(y1, m1, cal) - d1 + 1;
    for (let i = m1 + 1; i <= 12; i += 1) {
      no += getDaysInMonth(y1, i, cal);
    }

    // Get number of days till the required year
    for (let i = y1 + 1; i < y; i += 1) {
      no += getDaysInYear(i, cal);
    }

    // Get number of days till the required month
    for (let i = 1; i < m; i += 1) {
      no += getDaysInMonth(y, i, cal);
    }

    // Get number of days till the required date
    no += d;
    no -= 1;
  } else if (m !== m1) {
    // If the months differ
    // Get number of days 'no' till end of month
    no = getDaysInMonth(y, m1, cal) - d1 + 1;

    // Get number of days till the required month
    for (let i = m1 + 1; i < m; i += 1) {
      no += getDaysInMonth(y, i, cal);
    }

    // Get number of days till the required date
    no += d;
    no -= 1;
  } else {
    // Get number of days till the required date
    no = d - d1;
  }
  return sign * no;
}
function addYearsInDate(date, no) {
  const {
    cal,
    m
  } = date;
  const y = date.y + no;
  // Fix overflown day
  const d = Math.min(date.d, getDaysInMonth(y, m, cal));
  return prepareDate(y, m, d, cal);
}
function addMonthsInDate(date, no) {
  const {
    cal
  } = date;
  // NOTE: 'no' can be more than 12
  const y = date.y + Math.floor((date.m + no - 1) / 12);
  const m = (date.m + no - 1) % 12 + 1;
  // Fix overflown da)y
  const d = Math.min(date.d, getDaysInMonth(y, m, cal));
  return prepareDate(y, m, d, cal);
}
function addDaysInDate(date, no) {
  // return a clone of itself
  if (no === 0) {
    return prepareDate(date.y, date.m, date.d, date.cal);
  }

  // local variables that can be changed
  let {
    y,
    m,
    d
  } = date;
  const {
    cal
  } = date;

  // Get number of days 'n' till end of current year
  // 1. days remaining in current month
  let n = getDaysInMonth(y, m, cal) - d + 1;
  // 2. days remaining in following months
  for (let i = m + 1; i <= 12; i += 1) {
    n += getDaysInMonth(y, i, cal);
  }
  if (n <= no) {
    // If 'no' exceeds number of days in current year
    // Increment year
    y += 1;
    while (n + getDaysInYear(y, cal) <= no) {
      n += getDaysInYear(y, cal);
      y += 1;
    }
    // Increment month
    m = 1;
    while (n + getDaysInMonth(y, m, cal) <= no && m < 12) {
      n += getDaysInMonth(y, m, cal);
      m += 1;
    }
    // Increment day
    d = 1;
    d += no - n;
  } else {
    // Get number of days 'n' till end of month
    n = getDaysInMonth(y, m, cal) - d + 1;
    // If 'no' exceeds number of days in current month
    if (n <= no) {
      // Increment month
      m += 1;
      while (n + getDaysInMonth(y, m, cal) <= no) {
        n += getDaysInMonth(y, m, cal);
        m += 1;
      }
      // Increment day
      d = no - n + 1;
    } else {
      // Get day
      d += no;
    }
  }
  return prepareDate(y, m, d, cal);
}
function subtractDaysInDate(date, no) {
  if (no === 0) {
    return prepareDate(date.y, date.m, date.d, date.cal);
  }
  let {
    y,
    m,
    d
  } = date;
  const {
    cal
  } = date;

  // Get number of days 'n' till beginning of year
  // 1. get number of days elapsed in current month
  let n = d;
  // 2. get number of days elapsed in previous months
  for (let i = m - 1; i >= 1; i -= 1) {
    n += getDaysInMonth(y, i, cal);
  }

  // If 'no' exceeds number of days elapsed in current year
  if (n <= no) {
    // Decrement year
    y -= 1;
    while (n + getDaysInYear(y, cal) <= no) {
      n += getDaysInYear(y, cal);
      y -= 1;
    }
    // Decrement month
    m = 12;
    while (n + getDaysInMonth(y, m, cal) <= no && m > 1) {
      n += getDaysInMonth(y, m, cal);
      m -= 1;
    }
    // Decrement date
    d = getDaysInMonth(y, m, cal) - (no - n);
  } else {
    // Get number of days elapsed in current month
    n = d;
    // If 'no' exceeds number of days elapsed in current month
    if (n <= no) {
      // Decrement month
      m -= 1;
      while (n + getDaysInMonth(y, m, cal) <= no) {
        n += getDaysInMonth(y, m, cal);
        m -= 1;
      }
      // Decrement date
      d = getDaysInMonth(y, m, cal) - (no - n);
    } else {
      // Get date
      d -= no;
    }
  }
  return prepareDate(y, m, d, cal);
}
function getWeekFromDate(date) {
  const minDate = date.cal === 'AD' ? minAdDate : minBsDate;
  const totalDays = getDateDifference(date, minDate);
  return (totalDays + 6) % 7 + 1;
}
function convertDate(date, cal) {
  if (date.cal === cal) {
    return date;
  }
  const minDate = date.cal === 'AD' ? minAdDate : minBsDate;
  const totalDays = getDateDifference(date, minDate);
  const minDateForNewDate = cal === 'AD' ? minAdDate : minBsDate;
  const newDate = prepareDate(minDateForNewDate.y, minDateForNewDate.m, minDateForNewDate.d, cal);
  return addDaysInDate(newDate, totalDays);
}

exports.MIN_YEAR = MIN_YEAR;
exports._cs = _cs;
exports.addDaysInDate = addDaysInDate;
exports.addMonthsInDate = addMonthsInDate;
exports.addSeparator = addSeparator;
exports.addYearsInDate = addYearsInDate;
exports.basicTypes = basicTypes;
exports.bound = bound;
exports.breakFormat = breakFormat;
exports.bucket = bucket;
exports.camelToKebab = camelToKebab;
exports.camelToNormal = camelToNormal;
exports.camelToSnake = camelToSnake;
exports.capitalize = capitalize;
exports.caseInsensitiveSubmatch = caseInsensitiveSubmatch;
exports.checkVersion = checkVersion;
exports.compareBoolean = compareBoolean;
exports.compareDate = compareDate;
exports.compareLength = compareLength;
exports.compareNumber = compareNumber;
exports.compareString = compareString;
exports.compareStringAsNumber = compareStringAsNumber;
exports.compareStringByWordCount = compareStringByWordCount;
exports.compareStringSearch = compareStringSearch;
exports.convertDate = convertDate;
exports.decodeDate = decodeDate;
exports.difference = difference;
exports.doesObjectHaveNoData = doesObjectHaveNoData;
exports.encodeDate = encodeDate;
exports.findDifferenceInList = findDifferenceInList;
exports.formatDateToString = formatDateToString;
exports.formatPdfText = formatPdfText;
exports.formattedNormalize = formattedNormalize;
exports.getColorOnBgColor = getColorOnBgColor;
exports.getContrastYIQ = getContrastYIQ;
exports.getDate = getDate;
exports.getDateDifference = getDateDifference;
exports.getDateDifferenceHumanReadable = getDateDifferenceHumanReadable;
exports.getDaysInMonth = getDaysInMonth;
exports.getDaysInYear = getDaysInYear;
exports.getDefinedElementAround = getDefinedElementAround;
exports.getDifferenceInDays = getDifferenceInDays;
exports.getDuplicates = getDuplicates;
exports.getElementAround = getElementAround;
exports.getErrorForDateValues = getErrorForDateValues;
exports.getErrorForTimeValues = getErrorForTimeValues;
exports.getFirstKeyByValue = getFirstKeyByValue;
exports.getHashFromString = getHashFromString;
exports.getHexFromCode = getHexFromCode;
exports.getHexFromRgb = getHexFromRgb;
exports.getHexFromRgbRaw = getHexFromRgbRaw;
exports.getHexFromString = getHexFromString;
exports.getLinkedListNode = getLinkedListNode;
exports.getNumDaysInMonth = getNumDaysInMonth;
exports.getNumDaysInMonthX = getNumDaysInMonthX;
exports.getNumbers = getNumbers;
exports.getRandomFromList = getRandomFromList;
exports.getRatingForContentInString = getRatingForContentInString;
exports.getRgbFromHex = getRgbFromHex;
exports.getRgbRawFromHex = getRgbRawFromHex;
exports.getTrigramSimilarity = getTrigramSimilarity;
exports.getTrigrams = getTrigrams;
exports.getWeekFromDate = getWeekFromDate;
exports.interpolateRgb = interpolateRgb;
exports.intersection = intersection;
exports.isDateEqual = isDateEqual;
exports.isDateGreaterThan = isDateGreaterThan;
exports.isDateGreaterThanOrEqual = isDateGreaterThanOrEqual;
exports.isDateLessThan = isDateLessThan;
exports.isDateLessThanOrEqual = isDateLessThanOrEqual;
exports.isDateValuesComplete = isDateValuesComplete;
exports.isDefined = isDefined;
exports.isFalsy = isFalsy;
exports.isFalsyString = isFalsyString;
exports.isInteger = isInteger;
exports.isLeapYear = isLeapYear;
exports.isList = isList;
exports.isListEqual = isListEqual;
exports.isNaN = isNaN;
exports.isNotDefined = isNotDefined;
exports.isObject = isObject;
exports.isParamRequired = isParamRequired;
exports.isTimeValuesComplete = isTimeValuesComplete;
exports.isTruthy = isTruthy;
exports.isTruthyString = isTruthyString;
exports.isValidDate = isValidDate;
exports.isValidEmail = isValidEmail;
exports.isValidHexColor = isValidHexColor;
exports.isValidUrl = isValidUrl;
exports.listToGroupList = listToGroupList;
exports.listToMap = listToMap;
exports.mapToList = mapToList;
exports.mapToMap = mapToMap;
exports.max = max;
exports.mean = mean;
exports.median = median;
exports.min = min;
exports.modulo = modulo;
exports.noOp = noOp;
exports.normalize = normalize;
exports.padStart = padStart;
exports.pick = pick;
exports.populateFormat = populateFormat;
exports.prepareDate = prepareDate;
exports.randomString = randomString;
exports.removeKey = removeKey;
exports.resolve = resolve;
exports.reverseRoute = reverseRoute;
exports.splitInWhitespace = splitInWhitespace;
exports.subtractDaysInDate = subtractDaysInDate;
exports.sum = sum;
exports.toDateString = toDateString;
exports.trimWhitespace = trimWhitespace;
exports.typeOf = typeOf;
exports.union = union;
exports.unique = unique;
//# sourceMappingURL=index.cjs.map
