import { Maybe } from './declarations';
declare enum ValueType {
    time = "time",
    date = "date"
}
interface Value {
    type?: ValueType;
    value: string;
}
/**
 * Break timedate format into time unit, date unit and other unit.
 *
 * @remarks
 * The date unit are: yyyy, yy, MMM, MM, dd, EEE.
 * The time unit are: hh, mm, ss, aaa.
 * Anything can come in between date and time unit and they will be treated as separator.
 * @param format format for datetime
 */
export declare function breakFormat(format: string): Value[];
/**
 * Populate the format list with date
 *
 * @param formatList the list of datetime units
 * @param date
 */
export declare function populateFormat(formatList: Value[], date: Date): Value[];
/**
 * Format date
 *
 * @param date
 * @param format format for datetime
 * @returns formatted date
 */
export declare function formatDateToString(date: Date, format: string): string;
/**
 * Set hour, minute and second to zero in given datetime
 *
 * @param date timestamp or a date string
 * @returns timestamp with hour, minute and second set to zero
 */
export declare function getDate(datetime: string | number): number;
/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
export declare function getDifferenceInDays(a: string | number, b: string | number): number;
/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
export declare function getDateDifferenceHumanReadable(a: string, b: string): string;
/**
 * Get number of days in certain year and month
 *
 * @param year
 * @param month
 * @returns number of days
 *
 * @remarks
 * Month starts from 1 (not zero)
 */
export declare function getNumDaysInMonthX(year: Maybe<number>, month: Maybe<number>): number;
/**
 * Get number of days in certain year and month from a date
 *
 * @param date
 * @returns number of days
 */
export declare function getNumDaysInMonth(date: Maybe<Date>): number;
/**
 * Change date into 'yyyy-MM-dd' string
 *
 * @param date
 */
export declare function encodeDate(date: Date): string;
/**
 * Change timestamp or 'yyyy-MM-dd' string into date
 *
 * @param value timestamp or 'yyyy-MM-dd' string
 */
export declare function decodeDate(value: undefined | null): undefined;
export declare function decodeDate(value: string | number): Date;
export declare const MIN_YEAR = 1990;
interface Ymd {
    yearValue: number;
    monthValue: number;
    dayValue: number;
}
/**
 * Identify if ymd value is complete
 *
 * @param ymd
 */
export declare function isDateValuesComplete(ymd: Partial<Ymd>): ymd is Ymd;
/**
 * Identify problem with ymd value
 *
 * @param ymd
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
export declare function getErrorForDateValues(ymd: Partial<Ymd>): string | undefined;
interface Hms {
    hourValue: number;
    minuteValue: number;
    secondValue: number;
}
/**
 * Identify if hms value is complete
 *
 * @param hms
 */
export declare const isTimeValuesComplete: (val: Partial<Hms>) => val is Hms;
/**
 * Identify problem with hms value
 *
 * @param hms
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
export declare const getErrorForTimeValues: (val: Partial<Hms>) => string | undefined;
export {};
