import { padStart } from './string.js';
import { isTruthy, isFalsy } from './core.js';

// FIXME: do not use date library

const MONTHS = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
const DAYS = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
var ValueType = /*#__PURE__*/function (ValueType) {
  ValueType["time"] = "time";
  ValueType["date"] = "date";
  return ValueType;
}(ValueType || {});
function getStartAndEnd(format, matches) {
  let minStart;
  let maxEnd;
  matches.forEach(val => {
    const start = format.indexOf(val);
    if (start === -1) {
      return;
    }
    const end = start + val.length;
    if (minStart === undefined || start < minStart) {
      minStart = start;
    }
    if (maxEnd === undefined || end > maxEnd) {
      maxEnd = end;
    }
  });
  return {
    start: minStart,
    end: maxEnd
  };
}

/**
 * Break timedate format into time unit, date unit and other unit.
 *
 * @remarks
 * The date unit are: yyyy, yy, MMM, MM, dd, EEE.
 * The time unit are: hh, mm, ss, aaa.
 * Anything can come in between date and time unit and they will be treated as separator.
 * @param format format for datetime
 */
function breakFormat(format) {
  const {
    start: dateStart,
    end: dateEnd
  } = getStartAndEnd(format, ['yyyy', 'yy', 'MMM', 'MM', 'dd', 'EEE']);
  const {
    start: timeStart,
    end: timeEnd
  } = getStartAndEnd(format, ['hh', 'mm', 'ss', 'aaa']);
  const formatEnd = format.length;
  const getSubFormat = (start, end) => {
    if (start === undefined || end === undefined) {
      return '';
    }
    return format.substring(start, end);
  };
  if (dateStart === undefined && timeStart === undefined) {
    return [{
      value: format
    }].filter(a => a.value !== '');
  }
  if (dateStart === undefined) {
    return [{
      value: getSubFormat(0, timeStart)
    }, {
      type: ValueType.time,
      value: getSubFormat(timeStart, timeEnd)
    }, {
      value: getSubFormat(timeEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  if (timeStart === undefined) {
    return [{
      value: getSubFormat(0, dateStart)
    }, {
      type: ValueType.date,
      value: getSubFormat(dateStart, dateEnd)
    }, {
      value: getSubFormat(dateEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  if (dateStart < timeStart) {
    return [{
      value: getSubFormat(0, dateStart)
    }, {
      type: ValueType.date,
      value: getSubFormat(dateStart, dateEnd)
    }, {
      value: getSubFormat(dateEnd, timeStart)
    }, {
      type: ValueType.time,
      value: getSubFormat(timeStart, timeEnd)
    }, {
      value: getSubFormat(timeEnd, formatEnd)
    }].filter(a => a.value !== '');
  }
  return [{
    value: getSubFormat(0, timeStart)
  }, {
    type: ValueType.time,
    value: getSubFormat(timeStart, timeEnd)
  }, {
    value: getSubFormat(timeEnd, dateStart)
  }, {
    type: ValueType.date,
    value: getSubFormat(dateStart, dateEnd)
  }, {
    value: getSubFormat(dateEnd, formatEnd)
  }].filter(a => a.value !== '');
}

/**
 * Populate the format list with date
 *
 * @param formatList the list of datetime units
 * @param date
 */
function populateFormat(formatList, date) {
  return formatList.map(format => {
    if (format.type === ValueType.date) {
      const year = date.getFullYear();
      const month = date.getMonth();
      const day = date.getDate();
      const weekName = DAYS[date.getDay()];
      const newFormat = {
        ...format
      };
      newFormat.value = newFormat.value.replace('yyyy', String(year)).replace('yy', String(year % 100)).replace('MMM', MONTHS[month]).replace('MM', padStart(month + 1, 2)).replace('EEE', weekName).replace('dd', padStart(day, 2));
      return newFormat;
    }
    if (format.type === ValueType.time) {
      const ttIndex = format.value.indexOf('aaa');
      const originalHour = date.getHours();
      const hour = ttIndex >= 0 ? (originalHour - 1) % 12 + 1 : originalHour;
      const amPm = originalHour >= 12 ? 'PM' : 'AM';
      const minute = date.getMinutes();
      const second = date.getSeconds();
      const newFormat = {
        ...format
      };
      newFormat.value = newFormat.value.replace('hh', padStart(hour, 2)).replace('mm', padStart(minute, 2)).replace('ss', padStart(second, 2)).replace('aaa', amPm);
      return newFormat;
    }
    return format;
  });
}

/**
 * Format date
 *
 * @param date
 * @param format format for datetime
 * @returns formatted date
 */
function formatDateToString(date, format) {
  return populateFormat(breakFormat(format), date).map(e => e.value).join('');
}

/**
 * Set hour, minute and second to zero in given datetime
 *
 * @param date timestamp or a date string
 * @returns timestamp with hour, minute and second set to zero
 */
function getDate(datetime) {
  const today = new Date(datetime);
  today.setHours(0, 0, 0, 0);
  return today.getTime();
}

/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
function getDifferenceInDays(a, b) {
  const dateA = getDate(a);
  const dateB = getDate(b);
  return (dateA - dateB) / (1000 * 60 * 60 * 24);
}

/**
 * Get number of days betwen two datetime
 *
 * @param a timestamp or a date string
 * @param b timestamp or a date string
 * @returns number of days between two datetime
 */
function getDateDifferenceHumanReadable(a, b) {
  const diff = getDifferenceInDays(a, b);
  if (diff === 0) {
    return 'Today';
  }
  if (diff === -1) {
    return 'Yesterday';
  }
  if (diff === 1) {
    return 'Tomorrow';
  }
  if (diff < 0) {
    return `${-diff} days ago`;
  }
  return `After ${diff} days`;
}

/**
 * Get number of days in certain year and month
 *
 * @param year
 * @param month
 * @returns number of days
 *
 * @remarks
 * Month starts from 1 (not zero)
 */
function getNumDaysInMonthX(year, month) {
  if (isTruthy(year) && isTruthy(month)) {
    return new Date(year, month, 0).getDate();
  }
  return 32;
}

/**
 * Get number of days in certain year and month from a date
 *
 * @param date
 * @returns number of days
 */
function getNumDaysInMonth(date) {
  if (date) {
    return getNumDaysInMonthX(date.getFullYear(), date.getMonth() + 1);
  }
  return 32;
}

/**
 * Change date into 'yyyy-MM-dd' string
 *
 * @param date
 */
function encodeDate(date) {
  return formatDateToString(date, 'yyyy-MM-dd');
}

/**
 * Change timestamp or 'yyyy-MM-dd' string into date
 *
 * @param value timestamp or 'yyyy-MM-dd' string
 */

function decodeDate(value) {
  // Let's assume that the value is in local time zone

  if (!value) {
    return undefined;
  }

  // Check if value is timestamp number or ISO string with time information
  // In both case, new Date assumes local time zone
  if (typeof value === 'number' || value.includes('T')) {
    return new Date(value);
  }

  // In case of ISO string with no time information, new Date assumes
  // UTC timezone. So first split it manually, and feed them separately
  // so they will be processed as local time zone.
  const splits = value.split('-');
  return new Date(Number(splits[0]), Number(splits[1]) - 1, Number(splits[2]));
}
const MIN_YEAR = 1990;
/**
 * Identify if ymd value is complete
 *
 * @param ymd
 */
function isDateValuesComplete(ymd) {
  const {
    yearValue,
    monthValue,
    dayValue
  } = ymd;
  // Complete if all values are undefined or none are
  return isTruthy(dayValue) && isTruthy(monthValue) && isTruthy(yearValue) || isFalsy(dayValue) && isFalsy(monthValue) && isFalsy(yearValue);
}

/**
 * Identify problem with ymd value
 *
 * @param ymd
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
function getErrorForDateValues(ymd) {
  if (!isDateValuesComplete(ymd)) {
    return 'Date values incomplete';
  }
  const {
    yearValue,
    monthValue,
    dayValue
  } = ymd;
  if (yearValue < MIN_YEAR) {
    return `Year should be greater than or equal to ${MIN_YEAR}`;
  }
  if (monthValue < 1 || monthValue > 12) {
    return 'Month should be between 1 and 12';
  }
  const maxNumberOfDays = getNumDaysInMonthX(yearValue, monthValue);
  if (dayValue < 1 || dayValue > maxNumberOfDays) {
    return 'Invalid day for this month';
  }
  return undefined;
}
const MIN_HOUR = 0;
const MAX_HOUR = 23;
const MIN_MINUTE = 0;
const MAX_MINUTE = 59;
const MIN_SECOND = 0;
const MAX_SECOND = 59;
/**
 * Identify if hms value is complete
 *
 * @param hms
 */
const isTimeValuesComplete = val => {
  const {
    hourValue,
    minuteValue,
    secondValue
  } = val;
  // Complete if all values are undefined or none are
  return isTruthy(hourValue) && isTruthy(minuteValue) && isTruthy(secondValue) || isFalsy(hourValue) && isFalsy(minuteValue) && isFalsy(secondValue);
};

/**
 * Identify problem with hms value
 *
 * @param hms
 *
 * @remarks
 * The return value will be undefined if there are no problems.
 */
const getErrorForTimeValues = val => {
  if (!isTimeValuesComplete(val)) {
    return 'Time values incomplete';
  }
  const {
    hourValue,
    minuteValue,
    secondValue
  } = val;
  if (hourValue < MIN_HOUR || hourValue > MAX_HOUR) {
    return `Hour must be between ${MIN_HOUR} and ${MAX_HOUR}`;
  }
  if (minuteValue < MIN_MINUTE || minuteValue > MAX_MINUTE) {
    return `Minute must be between ${MIN_MINUTE} and ${MAX_MINUTE}`;
  }
  if (secondValue < MIN_SECOND || secondValue > MAX_SECOND) {
    return `Second must be between ${MIN_SECOND} and ${MAX_SECOND}`;
  }
  return undefined;
};

export { MIN_YEAR, breakFormat, decodeDate, encodeDate, formatDateToString, getDate, getDateDifferenceHumanReadable, getDifferenceInDays, getErrorForDateValues, getErrorForTimeValues, getNumDaysInMonth, getNumDaysInMonthX, isDateValuesComplete, isTimeValuesComplete, populateFormat };
//# sourceMappingURL=date.js.map
