import { isDefined } from './core.js';

/**
 * Get `a` modulo `b`
 *
 * @param a dividend
 * @param b divisor
 * @remarks
 * Return a positive modulo
 */
function modulo(a, b) {
  const value = a % b;
  if (value < 0) {
    return b + value;
  }
  // NOTE: -0 is not less than 0, so we need to take absolute value
  return Math.abs(value);
}

/**
 * Get sum of all elements in a list
 *
 * @param list of numbers
 *
 * @remarks
 * Return a positive modulo
 */
function sum(values) {
  return values.reduce((acc, b) => acc + b, 0);
}

/**
 * Get mean of all elements in a list
 *
 * @param list of numbers
 */
function mean(values) {
  if (values.length <= 0) {
    return 0;
  }
  return sum(values) / values.length;
}

/**
 * Get median of all elements in a list
 *
 * @param list of numbers
 */
function median(values) {
  if (values.length <= 0) {
    return undefined;
  }
  const sortedValues = [...values].sort();
  const centerIndex = Math.floor(sortedValues.length / 2);
  if (sortedValues.length % 2 !== 0) {
    return sortedValues[centerIndex];
  }
  const firstValue = sortedValues[centerIndex - 1];
  const secondValue = sortedValues[centerIndex];
  if (isDefined(firstValue) && isDefined(secondValue)) {
    return (firstValue + secondValue) / 2.0;
  }
  return firstValue ?? secondValue;
}

// min, max, output

/**
 * Get the bucket for a given number
 *
 * @param value
 * @param buckets
 */
function bucket(value, buckets) {
  const currentBucket = buckets.find(_ref => {
    let [min, max] = _ref;
    return value >= min && value < max;
  });
  return currentBucket && currentBucket[2];
}

/**
 * Get number bounded on two sides
 *
 * @param value
 * @param a upper or lower bound
 * @param b upper or lower bound
 */
function bound(value, a, b) {
  const min = Math.min(a, b);
  const max = Math.max(a, b);
  return Math.max(min, Math.min(max, value));
}

/**
 * Get normalized value for a number
 *
 * @param value
 * @param max max value
 * @param min min value
 */
function normalize(value, max, min) {
  return (value - min) / (max - min);
}

/**
 * Get list of numbers from `start` to `end`
 *
 * @param start number from where list starts
 * @param end number where list ends
 */
function getNumbers(start, end) {
  const list = [];
  for (let i = start; i < end; i += 1) {
    list.push(i);
  }
  return list;
}

export { bound, bucket, getNumbers, mean, median, modulo, normalize, sum };
//# sourceMappingURL=statistics.js.map
